from math import factorial

import matplotlib.pyplot as plt
import numpy as np

from numerical_analysis.dependencies import GeometricalPlace
from numerical_analysis.dependencies import Polynomial
from numerical_analysis.root_finding import newton_raphson


class Bezier(GeometricalPlace):

    def __init__(self, control_points: np.array, datatype=np.float):
        super().__init__()
        self.datatype = datatype
        self.cp = control_points
        self.n = len(self.cp) - 1
        self.m = self.matrix(self.n, self.datatype)
        self.c = self.polynomial_coefficients()
        self.polynomials = {"x": Polynomial(self.c[0]),
                            "y": Polynomial(self.c[1]),
                            "dx/dt": Polynomial(self.c[0]).derivative(),
                            "dy/dt": Polynomial(self.c[1]).derivative(),
                            "p": [Polynomial(self.m[i]) for i in range(self.n + 1)]}

    def polynomial_coefficients(self):
        c = np.empty([2, self.n + 1], dtype=self.datatype)
        for k in range(2):
            c[k] = np.matmul(np.transpose(self.m), self.cp[:, k])
        return c

    def y_x(self, x):
        t0 = newton_raphson(lambda t: self.x_t(t) - x, self.dx_dt, 0.5, 1e-14)
        return self.y_t(t0)

    def x_t(self, t):
        return self.polynomials["x"].value(t)

    def y_t(self, t):
        return self.polynomials["y"].value(t)

    def dx_dt(self, t):
        return self.polynomials["dx/dt"].value(t)

    def dy_dt(self, t):
        return self.polynomials["dy/dt"].value(t)

    def p_i(self, i, t):
        return self.polynomials["p"][i].value(t)

    def graph_cp(self):
        return [[self.cp[j, i] for j in range(len(self.cp))] for i in range(len(self.cp[0]))]

    def plot(self, dt):
        graph = self.graph(dt)
        # noinspection PyUnresolvedReferences
        plt.plot(graph[0], graph[1], self.graph_cp()[0], self.graph_cp()[1])
        plt.show()

    def modify_control_point(self, i, new_control_point):
        self.cp[i][0] = new_control_point[0]
        self.cp[i][1] = new_control_point[1]
        self.c = self.polynomial_coefficients()
        self.polynomials = {"x(t)": Polynomial(self.c[0]),
                            "y(t)": Polynomial(self.c[1]),
                            "dx/dt": Polynomial(self.c[0]).derivative(),
                            "dy/dt": Polynomial(self.c[1]).derivative(),
                            "p": [Polynomial(self.m[i]) for i in range(self.n + 1)]}

    def change_datatype(self, datatype):
        self.datatype = datatype
        self.cp = self.cp.astype(datatype)
        self.m = self.m.astype(datatype)
        self.c = self.c.astype(datatype)

    @classmethod
    def matrix(cls, n, datatype=np.int):
        return np.array([[((-1) ** (j - i) * factorial(n)) /
                          (factorial(i) * factorial(j - i) * factorial(n - j))
                          if j >= i else 0. for j in range(n + 1)] for i in range(n + 1)], dtype=datatype)



def secant(f, x0, x1, error=1e-15):
    fx0 = f(x0)
    fx1 = f(x1)
    while abs(fx1) > error:
        x2 = (x0 * fx1 - x1 * fx0) / (fx1 - fx0)
        x0, x1 = x1, x2
        fx0, fx1 = fx1, f(x2)
    return x1


def newton_raphson(f, df_dx, x0, error=1e-15):
    while abs(f(x0)) > error:
        x0 -= f(x0) / df_dx(x0)
    return x0


def newton_raphson_multiple_roots(f, df_dx, n, error=1e-15):

    def sigma(x):
        s = 0
        for root in roots:
            s += 1/(x - root)
        return s

    def nr_step(xold):
        return xold - f(xold) / (df_dx(xold) - f(xold)*sigma(xold))

    roots = []

    for i in range(n):

        xk_old = 2
        xk_new = nr_step(xk_old)
        ek = xk_new - xk_old
        xk_old = xk_new

        while abs(ek) > error:

            xk_new = nr_step(xk_old)
            ek = xk_new - xk_old
            xk_old = xk_new

        roots.append(xk_new)

    return sorted(roots)



import math

import matplotlib.pyplot as plt
from numpy import arange

from numerical_analysis.dependencies import Polynomial


class GeometricalPlace:

    def __init__(self):
        pass

    def x_t(self, t):
        return

    def y_t(self, t):
        return

    def point_t(self, t):
        return [self.x_t(t), self.y_t(t)]

    def graph(self, dt):
        graph = [[], []]
        for t in arange(0., 1. + dt, dt):
            for k in range(2):
                graph[k].append(self.point_t(t)[k])
        return graph

    def plot(self, dt):
        graph = self.graph(dt)
        plt.plot(graph[0], graph[1])
        plt.show()


class StraightLine(GeometricalPlace):

    def __init__(self, points):
        # points given in following format: [[t0, [x0, y0]], [t1, [x1, y1]]]
        self.points = points
        self.c = self.coefficients()
        super().__init__()

    def coefficients(self):
        t0 = self.points[0][0]
        t1 = self.points[1][0]

        x0 = self.points[0][1][0]
        x1 = self.points[1][1][0]

        y0 = self.points[0][1][1]
        y1 = self.points[1][1][1]

        a = (x1 - x0) / (t1 - t0)
        b = (x0 * t1 - x1 * t0) / (t1 - t0)
        c = (y1 - y0) / (t1 - t0)
        d = (y0 * t1 - y1 * t0) / (t1 - t0)

        return [[b, a], [d, c]]

    def x_t(self, t):
        return Polynomial(self.c[0]).value(t)

    def y_t(self, t):
        return Polynomial(self.c[1]).value(t)

    # noinspection PyMethodOverriding
    def graph(self, ta, tb):
        def invert_table(table):
            return [[table[j][i] for j in range(len(table))] for i in range(len(table[0]))]
        return invert_table([self.point_t(ta), self.point_t(tb)])

    def modify_points(self, points):
        self.points = points
        self.c = self.coefficients()


class Circle(GeometricalPlace):

    def __init__(self, center, radius):
        self.C = center
        self.R = radius
        super().__init__()

    def x_t(self, t):
        return self.R * math.cos(math.pi - 2 * math.pi * t) + self.C[0]

    def y_t(self, t):
        return self.R * math.sin(math.pi - 2 * math.pi * t) + self.C[1]



import numpy as np
from numerical_analysis.dependencies import Polynomial
from numerical_analysis.root_finding import newton_raphson_multiple_roots


def trapezoid(f, xs, xn, n):

    def dI():
        nonlocal xa, xb
        return (xb - xa) * (f(xa) + f(xb)) / 2

    I = 0
    dx = (xn - xs) / n

    for i in range(n):

        xa = xs + i * dx
        xb = xs + (i + 1) * dx

        I += dI()

    return I


def simpson1_3(f, xs, xn, n):

    def dI():
        nonlocal xa, xb

        x0 = xa
        x1 = (xa + xb) / 2
        x2 = xb

        return (xb - xa) * (f(x0) + 4 * f(x1) + f(x2)) / 6

    I = 0
    dx = (xn - xs) / n

    for i in range(n):

        xa = xs + i * dx
        xb = xs + (i + 1) * dx

        I += dI()

    return I


def simpson3_8(f, xs, xn, n):

    def dI():
        nonlocal xa, xb

        h = (xb - xa) / 3

        x0 = xa
        x1 = xa + h
        x2 = xb - h
        x3 = xb

        return (xb - xa) * (f(x0) + 3 * f(x1) + 3 * f(x2) + f(x3)) / 8

    I = 0
    dx = (xn - xs) / n

    for i in range(n):

        xa = xs + i * dx
        xb = xs + (i + 1) * dx

        I += dI()

    return I


def romberg(f, x0, xn, level):

    dx = (xn-x0)/2**(level-1)
    Irom = []
    xs = [x0 for i in range(level)]

    for i in range(level):
        Irom.append([])
        for j in range(level - i):
            Irom[i].append(0)

    for i in range(1, 2**(level-1) + 1):
        xi = x0 + i*dx
        for j in range(level):
            if i % 2**(level - (j+1)) == 0:
                Irom[j][0] += trapezoid(f, xs[j], xi, 1)
                xs[j] = xi

    for j in range(1, level):
        for i in range(level - j):
            Irom[i][j] = (4 ** j * Irom[i + 1][j - 1] - Irom[i][j - 1]) / (4 ** j - 1)

    return Irom[0][-1]


def gauss_legendre(f, a, b, n):

    p = Polynomial.legendre(n+1)

    x = newton_raphson_multiple_roots(p.value, p.derivative().value, p.n, 1e-14)
    L = Polynomial.lagrange(np.array(x))
    w = []

    for i in range(n+1):
        w.append(L[i].definite_integral(-1, 1))

    I = 0

    for i in range(n+1):
        I += w[i]*f((x[i]*(b-a)+b+a)/2)

    I *= (b-a)/2

    return I



import math
import numpy as np


class Polynomial:

    def __init__(self, coefficients: np.array):

        self.c = coefficients
        self.n = len(coefficients) - 1

    def value(self, x):
        return self.c.dot(np.array([x ** i for i in range(self.n + 1)]))

    def derivative(self):
        return Polynomial(np.delete(np.array([i for i in range(self.n + 1)]) * self.c, 0))

    def indefinite_integral(self):
        return Polynomial(np.insert(self.c / np.array([i + 1 for i in range(self.n + 1)]), 0, 0))

    def definite_integral(self, a, b):
        return self.indefinite_integral().value(b) - self.indefinite_integral().value(a)

    def graph(self, a, b, step):
        return [[x, self.value(x)] for x in np.arange(a, b + step, step)]

    @classmethod
    def product(cls, *input_polynomials):

        p = [polynomial.c for polynomial in input_polynomials]

        while len(p) > 1:

            n1 = len(p[-2]) - 1
            n2 = len(p[-1]) - 1

            p_temp = [0] * (n1 + n2 + 1)

            for i in range(n1 + 1):
                for j in range(n2 + 1):
                    p_temp[i + j] += p[-2][i] * p[-1][j]

            del p[-2:]

            p.append(p_temp)

        return cls(p[0])

    @classmethod
    def legendre(cls, n):

        c = np.array([0.] * (n + 1))

        for m in range(0, n // 2 + 1):
            c[n - 2 * m] = (-1) ** m * math.factorial(2 * n - 2 * m) / \
                           (2 ** n * math.factorial(m) * math.factorial(n - m) * math.factorial(n - 2 * m))

        return cls(c)

    @classmethod
    def lagrange(cls, x):

        n = x.size
        lagrange_polynomial = []

        for i in range(n):
            monomials = []
            for j in range(n):
                if j != i:
                    denominator = x[i] - x[j]
                    monomials.append(cls(np.array([-x[j] / denominator, 1 / denominator])))

            lagrange_polynomial.append(cls.product(*monomials))

        return lagrange_polynomial



import numpy as np
import matplotlib.pyplot as plt

from numerical_analysis.dependencies import GeometricalPlace
from numerical_analysis.splines import Bezier


class CompositeQuadraticBezier(GeometricalPlace):

    def __init__(self, control_points: np.array, datatype=np.float):
        super().__init__()
        self.datatype = datatype
        self.cp = control_points
        self.cp_enriched = control_points
        self.enrich_control_points()
        self.sectors = [Bezier(np.array([self.cp_enriched[2*i], self.cp_enriched[2*i + 1], self.cp_enriched[2*i + 2]]))
                        for i in range((len(self.cp_enriched)-1)//2)]

    def enrich_control_points(self):

        def mid_cp(cp0, cp1):
            return np.array([0.5 * (cp0[0] + cp1[0]), 0.5 * (cp0[1] + cp1[1])])

        self.cp_enriched = self.cp

        i = 1
        for j in range(len(self.cp) - 3):
            self.cp_enriched = np.insert(self.cp_enriched, i + 1,
                                         mid_cp(self.cp_enriched[i], self.cp_enriched[i + 1]), 0)
            i += 2

    def translate_t(self, t):
        if t != 1:
            i = int(t * len(self.sectors))
            u = t * len(self.sectors) - i
        else:
            i = len(self.sectors) - 1
            u = 1.
        return i, u

    def y_x(self, x):
        def detect_sector():
            i = 0
            while True:
                if self.sectors[i].x_t(1.) >= x:
                    return i
                i += 1
        return self.sectors[detect_sector()].y_x(x)

    def x_t(self, t):
        i, u = self.translate_t(t)
        return self.sectors[i].x_t(u)

    def y_t(self, t):
        i, u = self.translate_t(t)
        return self.sectors[i].y_t(u)

    def dx_dt(self, t):
        i, u = self.translate_t(t)
        return self.sectors[i].dx_dt(u)

    def dy_dt(self, t):
        i, u = self.translate_t(t)
        return self.sectors[i].dy_dt(u)

    def graph_cp(self):
        return [[self.cp[j, i] for j in range(len(self.cp))] for i in range(len(self.cp[0]))]

    def plot(self, dt, show_polyline=True, plot=True, export=False, filename="curve.png", title=None):
        plt.clf()
        if title:
            plt.title(title)
        graph = self.graph(dt)
        # noinspection PyUnresolvedReferences
        if show_polyline:
            plt.plot(graph[0], graph[1], "blue", self.graph_cp()[0], self.graph_cp()[1], "orange")
        else:
            plt.plot(graph[0], graph[1])

        if plot:
            plt.show()

        if export:
            plt.savefig(filename)

    def modify_control_point(self, i, new_control_point):
        self.cp[i][0] = new_control_point[0]
        self.cp[i][1] = new_control_point[1]
        self.cp_enriched = self.cp
        self.enrich_control_points()
        self.sectors = [Bezier(np.array([self.cp_enriched[2*i], self.cp_enriched[2*i + 1], self.cp_enriched[2*i + 2]]))
                        for i in range((len(self.cp_enriched)-1)//2)]



import numpy as np
import numerical_analysis as na


def f(x):
    return 3 * (x - 1) ** 2 - 2 * x


def df_dx(x):
    return 6 * x - 8


cp = np.array([[1., 3.],
               [2., 4.],
               [3., 5.],
               [4., 1.],
               [5., 1.]])

bezier = na.splines.Bezier(cp)
bezier.plot(0.01)

comp_bezier = na.splines.CompositeQuadraticBezier(cp)
comp_bezier.plot(0.01)

xa = 3
xb = 5
n = 1000

print(na.integration.trapezoid(f, xa, xb, n))
print(na.integration.simpson1_3(f, xa, xb, n))
print(na.integration.simpson3_8(f, xa, xb, n))
print(na.integration.romberg(f, xa, xb, 10))
print(na.integration.gauss_legendre(f, xa, xb, 11))

print(na.root_finding.secant(f, 0, 1))
print(na.root_finding.newton_raphson(f, df_dx, 0))



from setuptools import setup, find_packages

install_requires = [
    'numpy',
    'matplotlib'
]


setup(
    name="numerical_analysis",
    version="0.0.2",
    packages=find_packages(),

    author="IAGerogiannis",
    author_email='iagerogiannis@gmail.com',
    description="Numerical Analysis contains some basic numerical methods for solving equations, integrating "
                "functions and parameterizing curves with Bezier Polynomials.",
    url="https://github.com/iagerogiannis/numerical-analysis",
    install_requires=install_requires,
    license="BSD"
)
